package cz.fidentis.analyst.mesh.kdtree;
import cz.fidentis.analyst.kdtree.KdNode;
import cz.fidentis.analyst.kdtree.KdTree;
import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshFacetImpl;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import cz.fidentis.analyst.mesh.core.MeshPointImpl;
import org.junit.jupiter.api.Test;

import javax.vecmath.Vector3d;
import java.util.*;

import static org.junit.jupiter.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.assertTrue;


/**
 * @author Maria Kocurekova
 */

public class KdTreeTest {

    private Vector3d position = new Vector3d(0.1f, 0.5f, 0.7f);
    private Vector3d normalAndTextCoord = new Vector3d(0,0,0);
    

    private void testCorrectlyBuiltTree(KdTree tree) {
        if (tree.getRoot() == null) {
            return;
        }
        
        final Queue<KdNode> queue = new LinkedList<>();
        queue.add(tree.getRoot());
        
        while (!queue.isEmpty()) {
            final KdNode node = queue.poll();
            
            final KdNode lesser = node.getLesser();
            if (lesser != null) {
                switch (node.getDepth() % 3) {
                    case 0:
                        assertTrue(node.getLocation().x > lesser.getLocation().x);
                        break;
                    case 1:
                        assertTrue(node.getLocation().y > lesser.getLocation().y);
                        break;
                    default:
                        assertTrue(node.getLocation().z > lesser.getLocation().z);
                }
                queue.add(lesser);
            }
            
            final KdNode greater = node.getGreater();
            if (greater != null) {
                switch (node.getDepth() % 3) {
                    case 0:
                        assertTrue(node.getLocation().x <= greater.getLocation().x);
                        break;
                    case 1:
                        assertTrue(node.getLocation().y <= greater.getLocation().y);
                        break;
                    default:
                        assertTrue(node.getLocation().z <= greater.getLocation().z);
                }
                queue.add(greater);
            }
        }
    }
    
    @Test
    public void testCorrectBuild() {
        final Set<MeshPoint> points = new HashSet<>();
        points.add(new MeshPointImpl(new Vector3d(-331, 203, 320), normalAndTextCoord, normalAndTextCoord));
        points.add(new MeshPointImpl(new Vector3d(-371, -222, -111), normalAndTextCoord, normalAndTextCoord));
        points.add(new MeshPointImpl(new Vector3d(-223, 190, 113), normalAndTextCoord, normalAndTextCoord));
        points.add(new MeshPointImpl(new Vector3d(275, -414, -378), normalAndTextCoord, normalAndTextCoord));
        points.add(new MeshPointImpl(new Vector3d(-357, 98, -217), normalAndTextCoord, normalAndTextCoord));
        points.add(new MeshPointImpl(new Vector3d(297, 403, 299), normalAndTextCoord, normalAndTextCoord));
        points.add(new MeshPointImpl(new Vector3d(145, 252, -77), normalAndTextCoord, normalAndTextCoord));
        points.add(new MeshPointImpl(new Vector3d(319, 13, 87), normalAndTextCoord, normalAndTextCoord));
        points.add(new MeshPointImpl(new Vector3d(-284, 40, -231), normalAndTextCoord, normalAndTextCoord));
        
        final KdTree tree = new KdTree(points);
        testCorrectlyBuiltTree(tree);
    }
    
    @Test
    public void testCorrectBuildUnbalanced() {
        final Set<MeshPoint> treeNodes = new HashSet<>();
        treeNodes.add(new MeshPointImpl(new Vector3d( 1,  1,  1), normalAndTextCoord, normalAndTextCoord));
        treeNodes.add(new MeshPointImpl(new Vector3d(-1,  1,  1), normalAndTextCoord, normalAndTextCoord));
        treeNodes.add(new MeshPointImpl(new Vector3d( 1, -1,  1), normalAndTextCoord, normalAndTextCoord));
        treeNodes.add(new MeshPointImpl(new Vector3d(-1, -1,  1), normalAndTextCoord, normalAndTextCoord));
        treeNodes.add(new MeshPointImpl(new Vector3d( 1,  1, -1), normalAndTextCoord, normalAndTextCoord));
        treeNodes.add(new MeshPointImpl(new Vector3d(-1,  1, -1), normalAndTextCoord, normalAndTextCoord));
        treeNodes.add(new MeshPointImpl(new Vector3d( 1, -1, -1), normalAndTextCoord, normalAndTextCoord));
        treeNodes.add(new MeshPointImpl(new Vector3d(-1, -1, -1), normalAndTextCoord, normalAndTextCoord));
        treeNodes.add(new MeshPointImpl(new Vector3d( 1,  2,  4), normalAndTextCoord, normalAndTextCoord));
        treeNodes.add(new MeshPointImpl(new Vector3d( 1,  -7,  4), normalAndTextCoord, normalAndTextCoord));
        treeNodes.add(new MeshPointImpl(new Vector3d( 1,  5,  -1), normalAndTextCoord, normalAndTextCoord));
        treeNodes.add(new MeshPointImpl(new Vector3d( 1,  -2, -8), normalAndTextCoord, normalAndTextCoord));
        
        final KdTree kdTree = new KdTree(treeNodes);
        testCorrectlyBuiltTree(kdTree);
    }
    
    @Test
    public void testPut(){
        MeshPoint p = new MeshPointImpl(position,normalAndTextCoord,normalAndTextCoord);

        List<MeshFacet> facets = new LinkedList<>();
        MeshFacet facet = new MeshFacetImpl();
        facet.addVertex(p);
        facets.add(facet);
        KdTree tree = new KdTree(facets);
        testCorrectlyBuiltTree(tree);
        
        //System.out.println("AAA"+tree);

        assertTrue(tree.containsPoint(p));
    }

    @Test
    public void testPutNothing(){
        MeshPoint p = new MeshPointImpl(position,normalAndTextCoord,normalAndTextCoord);

        List<MeshFacet> facets2 = new LinkedList<>();
        KdTree tree = new KdTree(facets2);
        testCorrectlyBuiltTree(tree);

        assertFalse(tree.containsPoint(p));
    }

    /**
     * Returns the closest point
     *
     * @param points list of points
     * @param p point to which the closest point is searched
     * @return the closest point to p
     */
    private MeshPoint findClosestSequential(List<MeshPoint> points, MeshPoint p){
        if(points.isEmpty()){
            return null;
        }

        MeshPoint closest = points.get(0);
        double dis = closest.distanceTo(p);

        for(MeshPoint cl : points){
            if(cl.distanceTo(p) < dis){
                dis = cl.distanceTo(p);
                closest = cl;
            }
        }

        return closest;
    }

    @Test
    public void testFindClosestAlreadyIn(){
        List<MeshFacet> facets = new LinkedList<>();

        MeshFacet facet = new MeshFacetImpl();


        List<MeshPoint> points = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 0; i < 1000; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }
        facets.add(facet);
        KdTree tree = new KdTree(facets);
        testCorrectlyBuiltTree(tree);

        Random r = new Random();

        MeshPoint p = points.get(r.nextInt(points.size()));
        assertTrue(tree.containsPoint(p));
    }

    @Test
    public void testFindClosestAlreadyInTwoFacets(){
        List<MeshFacet> facets = new LinkedList<>();

        MeshFacet facet1 = new MeshFacetImpl();
        MeshFacet facet2 = new MeshFacetImpl();


        List<MeshPoint> points = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 0; i < 5; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet1.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }
        for(int i = 5; i < 10; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet2.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }
        facets.add(facet1);
        facets.add(facet2);
        KdTree tree = new KdTree(facets);
        testCorrectlyBuiltTree(tree);

        Random r = new Random();

        MeshPoint p = points.get(r.nextInt(points.size()));
        assertTrue(tree.containsPoint(p));
    }

    @Test
    public void testCountOfFacets(){
        List<MeshFacet> facets = new LinkedList<>();

        MeshFacet facet1 = new MeshFacetImpl();
        MeshFacet facet2 = new MeshFacetImpl();


        List<MeshPoint> points = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 0; i < 10; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet1.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }
        for(int i = 0; i < 10; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet2.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }
        facets.add(facet1);
        facets.add(facet2);
        KdTree tree = new KdTree(facets);
        testCorrectlyBuiltTree(tree);

        Random r = new Random();

        MeshPoint p = points.get(r.nextInt(points.size()));
        assertTrue(tree.containsPoint(p));
        KdNode node = tree.closestNode(p);
        assertNotNull(node);
        assertEquals(2, node.getFacets().size());
    }

    @Test
    public void testFindClosestNotInTreeTwoFacets(){
        List<MeshFacet> facets = new LinkedList<>();

        MeshFacet facet1 = new MeshFacetImpl();
        MeshFacet facet2 = new MeshFacetImpl();


        List<MeshPoint> points = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 0; i < 5; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet1.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }
        for(int i =5; i < 10; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet2.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }


        facets.add(facet1);
        facets.add(facet2);
        KdTree tree = new KdTree(facets);
        testCorrectlyBuiltTree(tree);

        MeshPoint p = new MeshPointImpl(new Vector3d(2.5f, 0.3f, 1.2f), normalAndTextCoord,normalAndTextCoord);
        KdNode node = tree.closestNode(p);
        assertNotNull(node);
        
        MeshPoint seq = findClosestSequential(points, p);
        assertNotNull(seq);
        
        assertEquals(node.getLocation(), seq.getPosition());
    }
    @Test
    public void testIndexFoundTwoFacetsNoDuplicates(){
        List<MeshFacet> facets = new LinkedList<>();

        MeshFacet facet1 = new MeshFacetImpl();
        MeshFacet facet2 = new MeshFacetImpl();


        List<MeshPoint> points = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 5; i < 20; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet1.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }
        for(int i =0; i < 5; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet2.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }


        facets.add(facet1);
        facets.add(facet2);
        KdTree tree = new KdTree(facets);
        testCorrectlyBuiltTree(tree);

        Random r = new Random();

        MeshPoint p = points.get(r.nextInt(points.size()));
        KdNode node = tree.closestNode(p);
        assertNotNull(node);
        assertEquals(node.getLocation(), p.getPosition());
        
        Map<MeshFacet, Integer> found = node.getFacets();
        Map.Entry<MeshFacet, Integer> entry = found.entrySet().iterator().next();
        assertEquals(entry.getKey().getVertex(entry.getValue()), p);        
    }

    @Test
    public void testIndexFoundTwoFacetsAllDuplicates(){
        List<MeshFacet> facets = new LinkedList<>();

        MeshFacet facet1 = new MeshFacetImpl();
        MeshFacet facet2 = new MeshFacetImpl();


        List<MeshPoint> points = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 0; i < 10; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet1.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }
        for(int i =0; i < 10; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet2.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }


        facets.add(facet1);
        facets.add(facet2);
        KdTree tree = new KdTree(facets);
        testCorrectlyBuiltTree(tree);

        Random r = new Random();

        MeshPoint p = points.get(r.nextInt(points.size()));
        KdNode node = tree.closestNode(p);
        assertNotNull(node);
        assertEquals(node.getLocation(), p.getPosition());
        
        Map<MeshFacet, Integer> found = node.getFacets();
        Map.Entry<MeshFacet, Integer> entry = found.entrySet().iterator().next();
        assertEquals(entry.getKey().getVertex(entry.getValue()), p);        

        MeshFacet second = (new ArrayList<>(found.keySet())).get(1);
        int secondIndex = (new ArrayList<>(found.values())).get(1);
        assertEquals(second.getVertex(secondIndex), p);
        
        assertEquals(0.0f, second.getVertex(secondIndex).distanceTo(entry.getKey().getVertex(entry.getValue())));
    }


    @Test
    public void testOneFacetNearNeighbourIndex(){
        MeshFacet facet1 = new MeshFacetImpl();


        List<MeshPoint> points = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 1; i < 20; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            facet1.addVertex(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }

        KdTree tree = new KdTree(facet1);
        testCorrectlyBuiltTree(tree);

        Random r = new Random();

        MeshPoint p = points.get(r.nextInt(points.size()));
        KdNode node = tree.closestNode(p);
        assertNotNull(node);
        assertEquals(node.getLocation(), p.getPosition());
        
        Map<MeshFacet, Integer> found = node.getFacets();
        Map.Entry<MeshFacet, Integer> entry = found.entrySet().iterator().next();
        assertEquals(entry.getKey().getVertex(entry.getValue()), p);
    }

    @Test
    public void testSetOfPointsNearNeighbourIndex(){
        Set<MeshPoint> points = new HashSet<>();
        List<MeshPoint> pointsList = new LinkedList<>();
        Vector3d positionOfPoints;

        for(int i = 1; i < 20; i++){
            positionOfPoints = new Vector3d(0.1f * i, 0.5f * i, 0.7f * i);
            points.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
            pointsList.add(new MeshPointImpl(positionOfPoints, normalAndTextCoord, normalAndTextCoord));
        }

        KdTree tree = new KdTree(points);
        testCorrectlyBuiltTree(tree);

        Random r = new Random();
        MeshPoint p = pointsList.get(r.nextInt(points.size()));
        KdNode node = tree.closestNode(p);
        assertNotNull(node);
        assertEquals(node.getLocation(), p.getPosition());
        
        Map<MeshFacet, Integer> found = node.getFacets();
        Map.Entry<MeshFacet, Integer> entry = found.entrySet().iterator().next();
        assertEquals(entry.getKey().getVertex(entry.getValue()), p);
    }
    
}


